local MAJOR = "LibScriptableDisplayPluginDPS-1.0" 
local MINOR = 15
local PluginDPS = LibStub:NewLibrary(MAJOR, MINOR)
if not PluginDPS then return end
local LibTimer = LibStub("LibScriptableDisplayTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires LibScriptableDisplayTimer-1.0")

local _G = _G
local frame = CreateFrame("Frame")
local data = {}
local update
local MAXRECORDS = 10

if not PluginDPS.__index then
	PluginDPS.__index = PluginDPS
end

-- Populate an environment with this plugin's fields
-- @usage :New(environment) 
-- @parma environment This will be the environment when setfenv is called.
-- @return A new plugin object, aka the environment, and the plugin object as second return
function PluginDPS:New(environment)

	environment.GetDPS = self.GetDPS
	environment.ResetDPS = self.ResetDPS
	environment.WipeDPS = self.WipeDPS
	
	self.timer = self.timer or LibTimer:New(MAJOR, 300, true, update)
	self.timer:Start()
	
	return environment
end

local function onEvent(self, event, ...)
	if PluginDPS[event] then
		PluginDPS[event](PluginDPS, ...)
	end
end

frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
frame:SetScript("OnEvent", onEvent)
frame:Show()

local GetTime = _G.GetTime
local UnitGUID = _G.UnitGUID

function PluginDPS.GetDPS(unit)
	local id = UnitGUID(unit)
	if not data[id] then return end
	local dps = data[id].damage / (GetTime() - data[id].startTime)
	return string.format("%.1f", dps), dps
end

function PluginDPS.ResetDPS(unit)
	local id = UnitGUID(unit)
	if not data[id] then return end
	data[id] = nil
end

function PluginDPS.WipeDPS()
	wipe(data)
end

local events = {
	SWING_DAMAGE = true,
	RANGE_DAMAGE = true,
	SPELL_DAMAGE = true,
	SPELL_PERIODIC_DAMAGE = true,
	DAMAGE_SHIELD = true,
	DAMAGE_SPLIT = true,
}

function PluginDPS:COMBAT_LOG_EVENT_UNFILTERED(_, eventType, id, _, _, _, _, _, spellID, _, _, damage)
	if not events[eventType] then return end
	if eventType == "SWING_DAMAGE" then
		damage = spellID
	end

	if not data[id] then
		data[id] = {}
		data[id].startTime = GetTime()
	end
	
	data[id].damage = (data[id].damage or 0) + damage
	data[id].lastUpdate = GetTime()
end

function update()
	for k, v in pairs(data) do
		local diff = GetTime() - (v.lastUpdate or 0)
		if v.lastUpdate and  diff > 5 then
			v.lastUpdate = GetTime()
			v.startTime = v.startTime + diff
		end
	end
end

